/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "liberty.h"
#include "hash.h"
#include "random.h"


/*
 * This file, together with engine/hash.h implements hashing of go positions
 * using a method known as Zobrist hashing.  See the Texinfo documentation
 * (Reading/Hashing) for more information.  
 */

static int is_initialized = 0;


#if 0
static Hashvalue
board_hash(Intersection board[MAX_BOARD][MAX_BOARD], int koi, int koj);

static void
board_to_position(Intersection board[MAX_BOARD][MAX_BOARD], 
		  int koi, int koj, Hashposition *pos);

static void 
oldhashdata_recalc(Hash_data *target, Intersection board[MAX_BOARD][MAX_BOARD],
		int koi, int koj);

#endif

static int
hashposition_compare(Hashposition *pos1, Hashposition *pos2);

/* Random values for the hash function.  For stones and ko position. */
static Hashvalue  white_hash[MAX_BOARD][MAX_BOARD];	
static Hashvalue  black_hash[MAX_BOARD][MAX_BOARD];	
static Hashvalue  ko_hash[MAX_BOARD][MAX_BOARD];


#define DYNAMICALLY_INITIALIZE_COMPACT 1
#if DYNAMICALLY_INITIALIZE_COMPACT
static Compacttype white_patterns[4 * sizeof(Compacttype)] = {0,};
static Compacttype black_patterns[4 * sizeof(Compacttype)] = {0,};

#else /* DYNAMICALLY_INITIALIZE_COMPACT */
#if SIZEOF_LONG >= 8  /* sizeof(Compacttype) >= 8 */
/* Bit patterns for white stones in the compact board representation. */
static Compacttype white_patterns[] = {
  0x0000000000000001ULL,0x0000000000000004ULL,0x0000000000000010ULL,0x0000000000000040ULL,
  0x0000000000000100ULL,0x0000000000000400ULL,0x0000000000001000ULL,0x0000000000004000ULL,
  0x0000000000010000ULL,0x0000000000040000ULL,0x0000000000100000ULL,0x0000000000400000ULL,
  0x0000000001000000ULL,0x0000000004000000ULL,0x0000000010000000ULL,0x0000000040000000ULL,
  0x0000000100000000ULL,0x0000000400000000ULL,0x0000001000000000ULL,0x0000004000000000ULL,
  0x0000010000000000ULL,0x0000040000000000ULL,0x0000100000000000ULL,0x0000400000000000ULL,
  0x0001000000000000ULL,0x0004000000000000ULL,0x0010000000000000ULL,0x0040000000000000ULL,
  0x0100000000000000ULL,0x0400000000000000ULL,0x1000000000000000ULL,0x4000000000000000ULL,
};

/* Bit patterns for black stones in the compact board representation. */
static Compacttype black_patterns[] = {
  0x0000000000000002ULL,0x0000000000000008ULL,0x0000000000000020ULL,0x0000000000000080ULL,
  0x0000000000000200ULL,0x0000000000000800ULL,0x0000000000002000ULL,0x0000000000008000ULL,
  0x0000000000020000ULL,0x0000000000080000ULL,0x0000000000200000ULL,0x0000000000800000ULL,
  0x0000000002000000ULL,0x0000000008000000ULL,0x0000000020000000ULL,0x0000000080000000ULL,
  0x0000000200000000ULL,0x0000000800000000ULL,0x0000002000000000ULL,0x0000008000000000ULL,
  0x0000020000000000ULL,0x0000080000000000ULL,0x0000200000000000ULL,0x0000800000000000ULL,
  0x0002000000000000ULL,0x0008000000000000ULL,0x0020000000000000ULL,0x0080000000000000ULL,
  0x0200000000000000ULL,0x0800000000000000ULL,0x2000000000000000ULL,0x8000000000000000ULL,
};
#else /* sizeof(Compacttype) < 8 */
/* Bit patterns for white stones in the compact board representation. */
static Compacttype white_patterns[] = {
  0x00000001, 0x00000004, 0x00000010, 0x00000040, 
  0x00000100, 0x00000400, 0x00001000, 0x00004000,
  0x00010000, 0x00040000, 0x00100000, 0x00400000,
  0x01000000, 0x04000000, 0x10000000, 0x40000000,
};

/* Bit patterns for black stones in the compact board representation. */
static Compacttype black_patterns[] = {
  0x00000002, 0x00000008, 0x00000020, 0x00000080, 
  0x00000200, 0x00000800, 0x00002000, 0x00008000,
  0x00020000, 0x00080000, 0x00200000, 0x00800000,
  0x02000000, 0x08000000, 0x20000000, 0x80000000,
};
#endif /* sizeof(Compacttype) < 8 */
#endif /* DYNAMICALLY_INITIALIZE_COMPACT */


/* Enable if you like to watch the hashing. */
#if 0
#define ECHOBEEP(_c) fputc(_c,stderr)
#else
#define ECHOBEEP(_c)
#endif


/* Get a random Hashvalue, where all bits are used. */
static Hashvalue
hash_rand(void)
{
  int i;
  Hashvalue h = 0;
  for (i=0; 32*i < (int) (CHAR_BIT*sizeof(Hashvalue)); i++)
    h |= (Hashvalue) gg_urand() << 32*i;

  return h;
}


/*
 * Initialize the entire hash system.
 */

static void
hash_init(void)
{
  int m, n;
  int x;
  struct gg_rand_state state;
  
  /* Since the hash initialization consumes a varying number of random
   * numbers depending on the size of the Hashvalue type, or none at
   * all if hashing is disabled, we save the state of the random
   * generator now and restore it afterwards.
   */
  gg_get_rand_state(&state);
  
#if TRACE_READ_RESULTS
  /* We need consistent hash values when this option is enabled. */
  gg_srand(1);
#endif
  
  for (m=0; m<MAX_BOARD; m++)
    for (n=0; n<MAX_BOARD; n++) {
      black_hash[m][n] = hash_rand();
      white_hash[m][n] = hash_rand();
      ko_hash[m][n]    = hash_rand();
    }
  
  gg_set_rand_state(&state);
  
#if DYNAMICALLY_INITIALIZE_COMPACT
  {
    Compacttype mask;

    for (x=0, mask=1; mask; x++, mask <<= 2) {
      white_patterns[x] = mask;
      black_patterns[x] = mask << 1;
    }
  }
#endif /* DYNAMICALLY_INITIALIZE_COMPACT */

  is_initialized = 1;
}


	/* AvK Removed these from hash.h: they are always
	** called together, so have been wrapped into 
	** hashdata_recalc() , which does the work without helpers.
	** They can be kept static and disappear.
	*/
#if 0
/*
 * Take a go position consisting of the board and a possible ko,
 * and generate a hash value from it.
 * 
 * See the Texinfo documentation (Reading/Hashing) for more information.
 */

static Hashvalue
board_hash(Intersection board[MAX_BOARD][MAX_BOARD], 
	   int koi, int koj)
{
  Hashvalue  hash;
  int           i;
  int           j;

  /* If the hash system is not initialized, do so now. */
  if (!is_initialized)
    hash_init();

  /* Get the hash value for all the stones on the board. */
  hash = 0;
  for (i = 0; i < board_size; ++i)
    for (j = 0; j < board_size; ++j) {
      switch (board[i][j]) {
      case WHITE:
	hash = hash ^ white_hash[i][j];
	break;
      case BLACK:
	hash = hash ^ black_hash[i][j];
	break;
      default:
	break;
      }
    }

  /* If there is a ko going on, take this into consideration too. */
  if (koi != -1) 
    hash ^= ko_hash[koi][koj];

  return hash;
}


/*
 * Take a go position consisting of the board and a possible ko,
 * and store this in a compact form into *pos.
 */

static void
board_to_position(Intersection board[MAX_BOARD][MAX_BOARD], 
		  int koi, int koj,  Hashposition *pos)
{
  int            index;
  int            subindex;
  Compacttype  bits;
  int            i, j;

  /* Initialize the compact board. */
  for (i = 0; i < (int) COMPACT_BOARD_SIZE; i++)
    pos->board[i] = 0;

  /* Put 16 locations into one long using 2 bits per location. */
  /* AvK 16 should read: 4* sizeof(Compacttype) long should read: Compacttype */
  index = 0;
  subindex = 0;
  bits = 0;
  for (i = 0; i < board_size; i++)
    for (j = 0; j < board_size; j++) {
      if (board[i][j] != EMPTY) {
	if (board[i][j] == WHITE)
	  bits |= white_patterns[subindex];
	else if (board[i][j] == BLACK)
	  bits |= black_patterns[subindex];
      }

      if (++subindex == POINTSPERCOMPACT) {
	pos->board[index++] = bits;
	bits = 0;
	subindex = 0;
      }
    }
  /* Store the last long into the compact board. 
     We only need to do this if we didn't just store it inside the loop. */
  if (subindex != 0)
    pos->board[index] = bits;

  /* The rest of the info, beside the board. */
  pos->ko_i    = koi;
  pos->ko_j    = koj;
}
#endif


/* ---------------------------------------------------------------- */


/* Return 0 if *pos1 == *pos2, otherwise return 1.
 * This adheres (almost) to the standard compare function semantics 
 * which are used e.g. by the comparison functions used in qsort().
 */

static int
hashposition_compare(Hashposition *pos1, Hashposition *pos2)
{
  int  i;

  /* We need only compare to board_size.  MAX_BOARD is not necessary. */
  for (i = 0; i < (int) (board_size * board_size / POINTSPERCOMPACT + 1); i++)
    if (pos1->board[i] != pos2->board[i]) {
      ECHOBEEP('A' + i);
      stats.hash_collisions++;
      return 1;
    }

  if (pos1->ko_i != pos2->ko_i || pos1->ko_j != pos2->ko_j) {
    ECHOBEEP('*');
    stats.hash_collisions++;
    return 1;
  }

  ECHOBEEP(' ');
  return 0;
}


/*
 * Dump an ASCII representation of the contents of a Hashposition onto
 * the FILE outfile. 
 */

void
hashposition_dump(Hashposition *pos, FILE *outfile)
{
  int  i;

  fprintf(outfile, "Board:  ");
  for (i = 0; i < (int) COMPACT_BOARD_SIZE; ++i)
     fprintf(outfile, " %lx", (unsigned long) pos->board[i]);

  if (pos->ko_i == -1)
    fprintf(outfile, "  No ko");
  else
    fprintf(outfile, "  Ko position: (%d, %d)", pos->ko_i, pos->ko_j);
}


/* ---------------------------------------------------------------- */


/* Calculate the compactboard and the hashvalue in one function.
 * They are always used together and it saves us a loop and a function call.
 * The USE_SHIFTING is just (CPU dependant?) optimization.
 * The theory behind it is that shifting is relaticely cheap.
 * Referencing a precomputed array uses the bus, and might consume 
 * a cacheslot. Just try.
 */

void 
hashdata_recalc(Hash_data *target, Intersection board[MAX_BOARD][MAX_BOARD],
		int koi, int koj)
{
#if 0
  oldhashdata_recalc(target, board, koi, koj);
#else

#define USE_SHIFTING 1
#if USE_SHIFTING
  unsigned int   index;
  int            i, j;
  Compacttype    bits;

  target->hashval = 0;
  bits = 1;
  index = 0;
  target->hashpos.board[index] = 0;
  for (i = 0; i < board_size; i++) {
    for (j = 0; j < board_size; j++) {
      switch (board[i][j]) {
      default:
      case EMPTY: 
	bits <<= 2;
	break;
      case WHITE:
	target->hashval ^= white_hash[i][j];
	target->hashpos.board[index] |= bits;
	bits <<= 2;
	break;
      case BLACK:
	target->hashval ^= black_hash[i][j];
	bits <<= 1;
	target->hashpos.board[index] |= bits;
	bits <<= 1;
	break;
      }
      if (!bits) { /* this means the 1bit fell off the left side */
	bits = 1;
	index += 1;
	if (index < COMPACT_BOARD_SIZE)
  	  target->hashpos.board[index] = 0;
      }
    }
  }
  /* This cleans up garbage bits at the(unused) end of the array */
  for ( ;++index < COMPACT_BOARD_SIZE; )
    target->hashpos.board[index] = 0;
#else /* USE_SHIFTING */
  int            index;
  int            subindex;
  int            i, j;
  Compacttype    bits;

  target->hashval = 0;
  index = 0;
  subindex = 0;
  bits = 0;
  for (i = 0; i < board_size; i++)
    for (j = 0; j < board_size; j++) {
      switch (board[i][j] ) {
      default:
      case EMPTY: 
	break;
      case WHITE:
	target->hashval ^= white_hash[i][j] ;
	bits |= white_patterns[subindex];
	break;
      case BLACK:
	target->hashval ^= black_hash[i][j] ;
	bits |= black_patterns[subindex];
	break;
      }

      if (++subindex == POINTSPERCOMPACT) {
	target->hashpos.board[index++] = bits;
	bits = 0;
	subindex = 0;
      }
    }
  
  if (subindex != 0)
    target->hashpos.board[index] = bits;
#endif /* USE_SHIFTING */

  if (koi >= 0) {
    target->hashval ^= ko_hash[koi][koj];
    target->hashpos.ko_i = koi;
    target->hashpos.ko_j = koj;
  } else {
    target->hashpos.ko_i = -1;
    target->hashpos.ko_j = -1;
  }
#endif /* 0 */
}


#if 0
static void 
oldhashdata_recalc(Hash_data *target, Intersection board[MAX_BOARD][MAX_BOARD],
		int koi, int koj)
{
  target->hashval = board_hash(board, koi, koj);
  board_to_position(board, koi, koj, &(target->hashpos));
}
#endif


/*
 * Set ko in the hash value and hash position.
 */

void
hashdata_set_ko(Hash_data *hd, int i, int j)
{
  hd->hashval ^= ko_hash[i][j];

  hd->hashpos.ko_i = i;
  hd->hashpos.ko_j = j;
}


/*
 * Remove any ko from the hash value and hash position.
 */

void
hashdata_remove_ko(Hash_data *hd)
{
  if (hd->hashpos.ko_i != -1) {
    hd->hashval ^= ko_hash[hd->hashpos.ko_i][hd->hashpos.ko_j];
    hd->hashpos.ko_i = -1;
    hd->hashpos.ko_j = -1;
  }
}


/*
 * Set or remove a stone of COLOR at (I, J) in a Hash_data.
 */

void
hashdata_invert_stone(Hash_data *hd, int i, int j, int color)
{
  int  index = (i * board_size + j) / POINTSPERCOMPACT;
  int  subindex = (i * board_size + j) % POINTSPERCOMPACT;

  if (color == BLACK) {
    hd->hashval ^= black_hash[i][j];
    hd->hashpos.board[index] ^= black_patterns[subindex];
  } else if (color == WHITE) {
    hd->hashval ^= white_hash[i][j];
    hd->hashpos.board[index] ^= white_patterns[subindex];
  }
}


/* ---------------------------------------------------------------- */


/*
 * Dump an ASCII representation of the contents of a Read_result onto
 * the FILE outfile. 
 */

void
read_result_dump(Read_result *result, FILE *outfile)
{
  fprintf(outfile, "Rutin %d, (%d, %d), djup: %d ",
	  ((result->routine_i_j_stackp) >> 24) & 0xff,
	  ((result->routine_i_j_stackp) >> 16) & 0xff,
	  ((result->routine_i_j_stackp) >> 8) & 0xff,
	  ((result->routine_i_j_stackp) >> 0) & 0xff);
  fprintf(outfile, "Resultat: %d, (%d, %d)\n",
	  ((result->result_ri_rj) >> 16) & 0xff,
	  ((result->result_ri_rj) >> 8) & 0xff,
	  ((result->result_ri_rj) >> 0) & 0xff);
}


/*
 * Dump an ASCII representation of the contents of a Hashnode onto
 * the FILE outfile. 
 */

void
hashnode_dump(Hashnode *node, FILE *outfile)
{
  Read_result  * result;

  /* Data about the node itself. */
  fprintf(outfile, "Hash value: %lx\n", (unsigned long) node->key.hashval);
  hashposition_dump(&(node->key.hashpos), outfile);

  for (result = node->results; result != NULL; result = result->next) {
    read_result_dump(result, outfile);
  }
  /* FIXME later: Dump contents of data also. */
}



/*
 * Initialize a hash table for a given total size and size of the
 * hash table.
 *
 * Return 0 if something went wrong.  Just now this means that there
 * wasn't enough memory available.
 */

int
hashtable_init(Hashtable *table, 
	       int tablesize, int num_nodes, int num_results)
{
  /* If the hash system is not initialized, do so now. */
  if (!is_initialized)
    hash_init();

  /* Allocate memory for the pointers in the hash table proper. */
  table->hashtablesize = tablesize;
  table->hashtable = (Hashnode **) malloc(tablesize * sizeof(Hashnode *));
  if (table->hashtable == NULL) {
    free(table);
    return 0;
  }

  /* Allocate memory for the nodes. */
  table->num_nodes = num_nodes;
  table->all_nodes = (Hashnode *) malloc(num_nodes * sizeof(Hashnode));
  if (table->all_nodes == NULL) {
    free(table->hashtable);
    free(table);
    return 0;
  }

  /* Allocate memory for the results. */
  table->num_results = num_results;
  table->all_results = (Read_result *) malloc(num_results 
					      * sizeof(Read_result));
  if (table->all_results == NULL) {
    free(table->hashtable);
    free(table->all_nodes);
    free(table);
    return 0;
  }

  /* Initialize the table and all nodes to the empty state . */
  hashtable_clear(table);

  return 1;
}


/*
 * Allocate a new hash table and return a pointer to it. 
 *
 * Return NULL if there is insufficient memory.
 */

Hashtable *
hashtable_new(int tablesize, int num_nodes, int num_results)
{
  Hashtable  * table;

  /* If the hash system is not initialized, do so now. */
  if (!is_initialized)
    hash_init();

  /* Allocate the hashtable struct. */
  table = (Hashtable *) malloc(sizeof(Hashtable));
  if (table == NULL)
    return NULL;

  /* Initialize the table. */
  if (!hashtable_init(table, tablesize, num_nodes, num_results)) {
    free(table);
    return NULL;
  }

  return table;
}


/*
 * Clear an existing hash table.  
 */

void
hashtable_clear(Hashtable *table)
{
  int bucket;
  if (!table)
    return;
  /* Initialize all hash buckets to the empty list. */
  for (bucket = 0; bucket < table->hashtablesize; ++bucket)
    table->hashtable[bucket] = NULL;

  table->free_node = 0;
  table->free_result = 0;
}


/*
 * Clear an existing hash table only if it happens to be full. By full
 * we mean that we are either out of positions or read results.
 */

void
hashtable_clear_if_full(Hashtable *table)
{
  if (table->free_node == table->num_nodes
      || table->free_result == table->num_results) {
    if (0)
      gprintf("Hashtable cleared because it was full.\n");
    hashtable_clear(table);
  }
}


/*
 * Enter a position with a given hash value into the table.  Return 
 * a pointer to the hash node where it was stored.  If it is already
 * there, don't enter it again, but return a pointer to the old one.
 */

Hashnode *
hashtable_enter_position(Hashtable *table, Hash_data *hd)
{
  Hashnode  * node;
  int         bucket;

  /* If the position is already in the table, return a pointer to it. */
  node = hashtable_search(table, hd);
  if (node != NULL) {
    return node;
  }

  /* If the table is full, return NULL */
  if (table->free_node == table->num_nodes)
    return NULL;

  /* It wasn't there and there is still room.  Allocate a new node for it... */
  node = &(table->all_nodes[table->free_node++]);
  node->key = *hd;
  node->results = NULL;

  /* ...and enter it into the table. */
  bucket = hd->hashval % table->hashtablesize;
  node->next = table->hashtable[bucket];
  table->hashtable[bucket] = node;

  ECHOBEEP('+');
  stats.position_entered++;
  return node;
}


/*
 * Currently we can't remove single nodes from the table.
 */

#if 0
void
hashtable_delete(Hashtable *table,
		 Intersection board[MAX_BOARD][MAX_BOARD])
{
}
#endif


/* 
 * Given a Hashposition and a Hash value, find the hashnode which contains
 * this very position with the given hash value.  
 *
 * We could compute the hash value within this functions, but later
 * when we have incremental calculation of the hash function, this 
 * would be dumb. So we demand the hash value from outside from the 
 * very beginning.
 */

Hashnode *
hashtable_search(Hashtable *table, Hash_data *hd)
{
  Hashnode     * node;
  int            bucket;

  bucket = hd->hashval % table->hashtablesize;
  for (node = table->hashtable[bucket]; node != NULL; node = node->next) {
    if (node->key.hashval != hd->hashval) {
      ECHOBEEP('#');
      continue;
    }
    if (hashposition_compare(&hd->hashpos, &node->key.hashpos) == 0)
      break;
    ECHOBEEP(7);
  }
  if (node)
    ECHOBEEP('!');
  else {
    ECHOBEEP('?');
  }
  return node;
}


/*
 * Dump an ASCII representation of the contents of a Hashtable onto
 * the FILE outfile. 
 */

void
hashtable_dump(Hashtable *table, FILE *outfile)
{
  int i;
  Hashnode * hn;

  /* Data about the table itself. */
  fprintf(outfile, "Dump of hashtable\n");
  fprintf(outfile, "Total size: %d\n", table->num_nodes);
  fprintf(outfile, "Size of hash table: %d\n", table->hashtablesize);
  fprintf(outfile, "Number of positions in table: %d\n", table->free_node);

  /* Data about the contents. */
  for (i = 0; i < table->hashtablesize; ++i) {
    fprintf(outfile, "Bucket %5d: ", i);
    hn = table->hashtable[i];
    if (hn == NULL)
      fprintf(outfile, "empty");
    else
      while (hn) {
	hashnode_dump(hn, outfile);
	hn = hn->next;
      }
    fprintf(outfile, "\n");
  }
}


/* 
 * Search the result list in a hash node for a particular result. This
 * result is from ROUTINE (e.g. readlad1) at (i, j) and reading depth
 * stackp.
 *
 * All these numbers must be unsigned, and 0<= x <= 255).
 */

Read_result *
hashnode_search(Hashnode *node, int routine, int i, int j)
{
  Read_result  * result;
  unsigned int   search_for;

  search_for = (((((routine << 8) | i) << 8) | j) << 8) | stackp;

  for (result = node->results; result != NULL; result = result->next) {
    if (result->routine_i_j_stackp == search_for)
      break;
    }

  return result;
}


/*
 * Enter a new Read_result into a Hashnode.
 * We already have the node, now we just want to enter the result itself.
 * We will fill in the result itself later, so we only need the routine
 * number for now.
 */

Read_result *
hashnode_new_result(Hashtable *table, Hashnode *node, 
		    int routine, int i, int j)
{
  Read_result  * result;

  /* If the table is full, return NULL */
  if (table->free_result == table->num_results)
    return NULL;

  /* There is still room. Allocate a new node for it... */
  result = &(table->all_results[table->free_result++]);

  /* ...and enter it into the table. */
  result->next = node->results;
  node->results = result;

  /* Now, put the routine number into it. */
  result->routine_i_j_stackp = ((((((routine << 8) | i) << 8) | j) << 8)
				| stackp);

  stats.read_result_entered++;
  return result;

}


/*
 * Compare two Hash_data, if different: dump an ASCII representation 
 * of the differences to stderr.
 * return is the same as for hashposition_compare()
 */

int
hashdata_diff_dump(Hash_data *key1, Hash_data *key2)
{
  int retval;
  int pos, i;
  int count1[4], count2[4];
  static const char letter[] = "abcdefghjklmnopqrstuvwxyz";
  static const char *hashcolors[] = {"Empty", "White", "Black", "Grey!"};

  ECHOBEEP(':');
  retval = hashdata_compare(key1, key2);
  if (retval == 0)
    return retval;

  for (i = 0; i < 4; i++) {
    count1[i] = 0;
    count2[i] = 0;
  }

  fprintf(stderr, "Differences: ");
  for (i = 0; i < COMPACT_BOARD_SIZE; i++) {
    if (key1->hashpos.board[i] != key2->hashpos.board[i])
      fprintf(stderr, "\nSlot %d: (%lx <==> %lx)" , i,
	      (unsigned long) key1->hashpos.board[i],
	      (unsigned long) key2->hashpos.board[i]);
    
    for (pos = 0; pos < POINTSPERCOMPACT; pos++) {
      unsigned int u1, u2;
      int xx, yy, zz;

      u1 = (key1->hashpos.board[i] >> (2*pos) ) & 3ULL;
      u2 = (key2->hashpos.board[i] >> (2*pos) ) & 3ULL;
      count1[u1] += 1;
      count2[u2] += 1;
      if (u1 == u2)
	continue;
      
      zz = (i * POINTSPERCOMPACT) + pos;
      xx = zz / MAX_BOARD;
      yy = zz % MAX_BOARD;
      fprintf(stderr, "\n#%2d: [%c%d] %s<==>%s", pos, letter[xx], yy,
	      hashcolors[u1], hashcolors[u2]);
    }
  }

  if (key1->hashpos.ko_i == -1 && key2->hashpos.ko_i == -1)
    fprintf(stderr, "\nNo ko\n");
  else if (key1->hashpos.ko_i == key2->hashpos.ko_i
	   && key1->hashpos.ko_j == key2->hashpos.ko_j )
    fprintf(stderr, "\nEqual Ko position:[%c%d]\n",
	    letter[key1->hashpos.ko_i], key1->hashpos.ko_j);
  else
    fprintf(stderr, "\nDifferent Ko position:[%c%d] <==> [%c%d]\n",
	    letter[key1->hashpos.ko_i], key1->hashpos.ko_j,
	    letter[key2->hashpos.ko_i], key2->hashpos.ko_j);

  fprintf(stderr, "Total [%d,%d,%d,%d]",
	  count1[0], count1[1], count1[2], count1[3]);
  fprintf(stderr, " <==> [%d,%d,%d,%d]\n",
	  count2[0], count2[1], count2[2], count2[3]);

  return retval;
}

int
hashdata_compare(Hash_data *key1, Hash_data *key2)
{
  int  rc;

  ECHOBEEP('<');
  rc = (key1->hashval == key2->hashval) ? 0 : 2;
  if (rc == 0)
    rc = hashposition_compare(&key1->hashpos, &key2->hashpos);
  ECHOBEEP('>');

  return rc;
}

/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
