/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include <string.h>
#include <assert.h>

#include "liberty.h"


/*
 * Put the values in POS into the global variables which is the
 * equivalent of the position.
 */

static void
position_to_globals(Position *pos)
{
  board_size = pos->boardsize;
  setup_board(pos->board, pos->ko_i, pos->ko_j,
	      pos->white_captured, pos->black_captured);
}


/*
 * Get the values in POS from the global variables which is the
 * equivalent of the position.
 */

static void
globals_to_position(Position *pos)
{
  pos->boardsize = board_size;
  memcpy(pos->board, p, sizeof(p));
  pos->ko_i = ko_i;
  pos->ko_j = ko_j;

  pos->white_captured = white_captured;
  pos->black_captured = black_captured;
}


/* ---------------------------------------------------------------- */


/*
 * Clear a position.
 */

void
position_clear(Position *pos, int boardsize)
{
  assert(MIN_BOARD <= boardsize && boardsize <= MAX_BOARD);
  pos->boardsize = boardsize;
  memset(pos->board, EMPTY, sizeof(pos->board));
  pos->ko_i = -1;
  pos->ko_j = -1;

  pos->white_captured = 0;
  pos->black_captured = 0;
}


void
position_copy(Position *to, Position *from)
{
  memcpy(to, from, sizeof(Position));
}


void
position_add_stone(Position *pos, int i, int j, int color)
{
  position_to_globals(pos);
  add_stone(i, j, color);
  globals_to_position(pos);
}


void
position_remove_stone(Position *pos, int i, int j)
{
  position_to_globals(pos);
  remove_stone(i, j);
  globals_to_position(pos);
}


void
position_play_move(Position *pos, int i, int j, int color)
{
  position_to_globals(pos);
  play_move(i, j, color);
  globals_to_position(pos);
}


int
position_legal(Position *pos, int i, int j, int color)
{
  int  retval;

  position_to_globals(pos);
  retval = legal(i, j, color);

  return retval;
}


int
position_placehand(Position *pos, int handicap)
{
  int  retval;

  position_to_globals(pos);
  retval = placehand(handicap);
  globals_to_position(pos);

  return retval;
}


/* FIXME: Recordhand should be moved from sethand and this function
          should either be removed or used differently. */

void
position_recordhand(Position *pos, int change_sgf_tree)
{
  position_to_globals(pos);
  recordhand(change_sgf_tree);
}


int
position_sethand(Position *pos, int handicap, int change_sgf_tree)
{
  int  retval;

  position_to_globals(pos);
  retval = sethand(handicap, change_sgf_tree);
  globals_to_position(pos);

  return retval;
}


int
position_genmove(Position *pos, int *i, int *j, int color, int move_number)
{
  movenum = move_number;
  position_to_globals(pos);
  return genmove(i, j, color);
}

int
position_attack(Position *pos, int m, int n, int *i, int *j)
{
  position_to_globals(pos);
  return attack(m, n, i, j);
}


int
position_find_defense(Position *pos, int m, int n, int *i, int *j)
{
  position_to_globals(pos);
  return find_defense(m, n, i, j);
}


void
position_who_wins(Position *pos, int color, float komi, FILE *outfile)
{
  position_to_globals(pos);
  who_wins(color, komi, outfile);
}


void
position_evaluate_territory(Position *pos, int *white, int *black)
{
  position_to_globals(pos);
  evaluate_territory(white, black);
}


/* ================================================================ */
/*                             Gameinfo                             */
/* ================================================================ */


/*
 * Initialize the structure.
 */

void
ginfo_clear(Gameinfo *ginfo, int boardsize)
{
  ginfo->handicap = 0;
  ginfo->komi = 5.5;
  
  ginfo->move_number = 0;
  position_clear(&(ginfo->position), boardsize);
  ginfo->to_move = BLACK;

  ginfo->computer_player = WHITE;
}


/*
 * Print a gameinfo.
 */

void
ginfo_print(Gameinfo *ginfo)
{
  printf("Board Size:   %d\n", ginfo->position.boardsize);
  printf("Handicap      %d\n", ginfo->handicap);
  printf("Komi:         %.1f\n", ginfo->komi);
  printf("Move Number:  %d\n", ginfo->move_number);

  printf("To Move:      ");
  if (ginfo->to_move == WHITE)
    printf("White\n");
  else if (ginfo->to_move == BLACK)
    printf("Black\n");
  else
    printf("Nobody\n");

  if (ginfo->computer_player == WHITE)
    printf("White\n");
  else if (ginfo->computer_player == BLACK)
    printf("Black\n");
  else if (ginfo->computer_player == EMPTY)
    printf("Both (solo)\n");
  else
    printf("Nobody\n");
}



/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
