/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




#ifndef _LIBERTY_H_
#define _LIBERTY_H_

#include "gnugo.h"

/* aix defines abs in stdlib.h */

#ifdef abs
#undef abs
#endif

/* MS-VC defines max and min */

#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif


#define abs(x) ((x) < 0 ? -(x) : (x))
#define min(a,b) ((a)<(b) ? (a) : (b))
#define max(a,b) ((a)<(b) ? (b) : (a))

/* not sure if this is the best way of doing this, but... */
#define UNUSED(x)  x=x

/* ================================================================ */
/*                           public variables                       */
/* ================================================================ */


/* A few variables below are of types defined in hash.h. */
#include "hash.h"


/* other modules get read-only access to these variables */
extern Hash_data    hashdata;
extern Hashtable  * movehash;




/* ================================================================ */


#define FALSE_EYE          1
#define HALF_EYE           2
#define INHIBIT_CONNECTION 4


/* A string with n stones can have at most 2(n+1) liberties. From this
 * follows that an upper bound on the number of liberties of a string
 * on a board of size N^2 is 2/3 (N^2+1).
 */
#define MAXLIBS   (2*(MAX_BOARD*MAX_BOARD + 1)/3)

/* This is an upper bound of the number of strings that can exist on
 * the board simultaneously.
 */
#define MAX_STRINGS (2 * MAX_BOARD * MAX_BOARD / 3)

#define MAXSTACK  160
#define MAXCHAIN  160


/* board utility functions */

void find_origin(int i, int j, int *origini, int *originj);
void chainlinks(int m, int n, int *adj, int adji[MAXCHAIN],
                int adjj[MAXCHAIN], int adjsize[MAXCHAIN],
                int adjlib[MAXCHAIN]);

/* This is increased by one anytime a move is played or the board cleared. */
int position_number;

/* Count and/or find liberties at (i, j). */
int countlib(int m, int n);
int findlib(int m, int n, int maxlib, int *libi, int *libj);
int approxlib(int m, int n, int color, int maxlib, int *libi, int *libj);


/* Play at (m, n) and then count the liberties. */
int accurate_approxlib(int m, int n, int color, int maxlib,
		       int *libi, int *libj);

/* Count the number of stones in a string. */
int countsize(int m, int n);
  
void dump_stack(void);

/* Is this point inside the board? */
#define ON_BOARD(i, j) ((i)>=0 && (j)>=0 && (i)<board_size && (j)<board_size)

/* Coordinates for the eight directions, ordered
 * south, west, north, east, southwest, northwest, northeast, southeast.
 * Defined in utils.c.
 */
extern int deltai[8]; /* = { 1,  0, -1,  0,  1, -1, -1, 1}; */
extern int deltaj[8]; /* = { 0, -1,  0,  1, -1, -1,  1, 1}; */


struct pattern; /* keep gcc happy */
struct pattern_db; /* keep gcc happy */
struct fullboard_pattern; /* keep gcc happy */

/*
 * try to match a pattern in the database to the board. Callback for each match */
typedef void (*matchpat_callback_fn_ptr)(int m, int n, int color,
					 struct pattern *, int rotation,
					 void *data);
typedef void (*fullboard_matchpat_callback_fn_ptr)(int ti, int tj,
						   struct fullboard_pattern *,
						   int rotation);
void matchpat(int m, int n, matchpat_callback_fn_ptr callback, int color,
	      struct pattern_db *pdb, void *callback_data,
	      char goal[MAX_BOARD][MAX_BOARD]);
void fullboard_matchpat(fullboard_matchpat_callback_fn_ptr callback, int color,
			struct fullboard_pattern *pattern);

int find_defense(int m, int n, int *i, int *j);
int attack(int m, int n, int *i, int *j);

/* Should these be static? */
int defend1(int si, int sj, int *i, int *j);
int defend2(int si, int sj, int *i, int *j);
int defend3(int si, int sj, int *i, int *j);
int defend4(int si, int sj, int *i, int *j);
int attack1(int si, int sj, int *i, int *j);
int attack2(int si, int sj, int *i, int *j);
int attack3(int i, int j, int *ti, int *tj);
int attack4(int i, int j, int *ti, int *tj);
int break_chain(int si, int sj, int *i, int *j, int *k, int *l);
int break_chain2(int si, int sj, int *i, int *j);

int double_atari_chain2(int si, int sj, int *i, int *j);
int find_cap2(int m, int n, int *i, int *j);
int is_ko(int i, int j, int color);
int singleton(int i, int j);
void mark_string(int i, int j, char mx[MAX_BOARD][MAX_BOARD], char mark);
int move_in_stack(int m, int n, int cutoff);
int vital_chain(int m, int n);
int confirm_safety(int i, int j, int color, int size, int *di, int *dj);
void modify_depth_values(int n);
void increase_depth_values(void);
void decrease_depth_values(void);
void set_temporary_depth_values(int d, int b, int f, int k);
void restore_depth_values(void);
int safe_move(int i, int j, int color);
int naive_ladder(int si, int sj, int *i, int *j);
int safe_stone(int i, int j);
void join_dragons(int ai, int aj, int bi, int bj);
void compute_escape_potential(void);
int find_border(int m, int n, int *edge, int *size);
void propagate_worm(int m, int n);
int strategic_distance_to(int color, int i, int j);
int distance_to(int color, int i, int j);
void transform(int i, int j, int *ti, int *tj, int trans);
void offset(int i, int j, int basei, int basej, int *ti, int *tj, int trans);
void find_cuts(void);
void find_connections(void);
void modify_eye_spaces1(void);
void modify_eye_spaces2(void);
void endgame(void);

/* functions to add (or remove) move reasons */
void clear_move_reasons(void);
void add_lunch(int ai, int aj, int bi, int bj);
void remove_lunch(int ai, int aj, int bi, int bj);
void add_defense_move(int ti, int tj, int ai, int aj);
void remove_defense_move(int ti, int tj, int ai, int aj);
int defense_move_known(int ti, int tj, int ai, int aj);
void add_attack_move(int ti, int tj, int ai, int aj);
void remove_attack_move(int ti, int tj, int ai, int aj);
int attack_move_known(int ti, int tj, int ai, int aj);
void add_connection_move(int ti, int tj, int ai, int aj, int bi, int bj);
void add_cut_move(int ti, int tj, int ai, int aj, int bi, int bj);
void add_antisuji_move(int ti, int tj);
void add_semeai_move(int ti, int tj, int ai, int aj, int bi, int bj);
void add_mouse_attack_move(int ti, int tj, int ai, int aj);
void add_mouse_defense_move(int ti, int tj, int ai, int aj);
void mouse_find_lunches(char goal[MAX_BOARD][MAX_BOARD], int color);
int mouse_lively(int i, int j);
void add_vital_eye_move(int ti, int tj, int ai, int aj, int color);
void add_attack_either_move(int ti, int tj, int ai, int aj, int bi, int bj);
void add_defend_both_move(int ti, int tj, int ai, int aj, int bi, int bj);
void add_block_territory_move(int ti, int tj);
void add_expand_territory_move(int ti, int tj);
void add_expand_moyo_move(int ti, int tj);
void add_strategical_attack_move(int ti, int tj, int ai, int aj);
void add_strategical_defense_move(int ti, int tj, int ai, int aj);
void set_minimum_move_value(int ti, int tj, float value);
void set_maximum_move_value(int ti, int tj, float value);
void set_minimum_territorial_value(int ti, int tj, float value);
void set_maximum_territorial_value(int ti, int tj, float value);
void add_shape_value(int ti, int tj, float value);
void add_followup_value(int ti, int tj, float value);
void mouse_reasons(int color);
int sniff_lunch(int i, int j, int *max, int *min);
int unconditional_life(int wormi[MAX_STRINGS], int wormj[MAX_STRINGS],
		       int color);

int sethand(int handicap, int change_sgf_tree); /* place and record */
int placehand(int handicap);          /* place stones on board only */
void recordhand(int change_sgf_tree); /* record board's stones in sgf */
                                      /* tree and output */

/* various different strategies for finding a move */

void fuseki(int color);
void semeai(int color);
void small_semeai(void);
void small_semeai_analyzer(int, int, int, int);
void shapes(int color);
void endgame_shapes(int color);

int review_move_reasons(int *i, int *j, float *val, int color);
int fill_liberty(int *i, int *j, int color);
int revise_semeai(int color);
int find_lunch(int m, int n, int *wi, int *wj, int *ai, int *aj);
int find_defense(int m, int n, int *ti, int *tj);
int mouse_attack(int m, int n, int *i, int *j, char goal[MAX_BOARD][MAX_BOARD]);
int mouse_defend(int m, int n, int *i, int *j, char goal[MAX_BOARD][MAX_BOARD]);
int mouse_effective_defense(int m, int n, int ti, int tj);
void change_defense(int, int, int, int, int);
void change_attack(int, int, int, int, int);
int does_attack(int, int, int, int);
int does_defend(int, int, int, int);
int double_atari(int m, int n, int color);
int attack_either(int ai, int aj, int bi, int bj);
int defend_both(int ai, int aj, int bi, int bj);
int break_through(int ai, int aj, int bi, int bj, int ci, int cj);
int play_attack_defend_n(int color, int do_attack, int num_moves, ...);
int play_attack_defend2_n(int color, int do_attack, int num_moves, ...);
int play_break_through_n(int color, int num_moves, ...);
int cut_possible(int i, int j, int color);

int liberty_of(int i, int j, int m, int n);
int defend_against(int ti, int tj, int color, int ai, int aj);

/* moyo.c */
int area_color(int m,int n);
int moyo_color(int m,int n);
int terri_color(int m,int n);
int delta_moyo(int ti, int tj,int color);
int delta_moyo_simple(int ti, int tj,int color);
int delta_terri(int ti, int tj,int color);
int diff_moyo(int ti, int tj, int color);
int diff_terri(int ti, int tj, int color);
int area_stone(int m, int n);
int area_space(int m, int n);

/* Printmoyo values, specified by -m flag. */
#define PRINTMOYO_TERRITORY         0x01
#define PRINTMOYO_DELTA_TERRITORY   0x02
#define PRINTMOYO_MOYO              0x04
#define PRINTMOYO_DELTA_MOYO        0x08
#define PRINTMOYO_AREA              0x10
#define PRINTMOYO_LIST_AREAS        0x20
#define PRINTMOYO_METACONNECT       0x40
/* The following have been borrowed by the influence functions below. */
#define PRINTMOYO_INITIAL_INFLUENCE 0x80
#define PRINTMOYO_PRINT_INFLUENCE   0x100
#define PRINTMOYO_NUMERIC_INFLUENCE 0x200

/* Influence functions. */
void compute_initial_influence(int color, int dragons_known);
int influence_delta_territory(int m, int n, int color);
int influence_delta_moyo(int m, int n, int color);
int influence_delta_strict_moyo(int m, int n, int color);
int influence_delta_area(int m, int n, int color);
int influence_delta_strict_area(int m, int n, int color);
int influence_territory_color(int m, int n);
int influence_moyo_color(int m, int n);
int influence_area_color(int m, int n);

/* Eye space functions. */
int eye_space(int i, int j);
int proper_eye_space(int i, int j);
int marginal_eye_space(int i, int j);
int halfeye(int i, int j);
void make_proper_eye_space(int i, int j, int color);
void remove_half_eye(int i, int j, int color);
void remove_eyepoint(int i, int j, int color);


/* debugging support */
void move_considered(int i, int j, int value);


/* SGF routines for debugging purposes in sgffile.c */
int  sgffile_write_line(const char *, ...);
void sgffile_dragon_status(int, int, int );
void goaldump(char goal[MAX_BOARD][MAX_BOARD]);


/* ================================================================ */
/*                         global variables                         */
/* ================================================================ */

/* The board and the other parameters deciding the current position. */
extern int           board_size;	     /* board size (usually 19) */
extern Intersection  p[MAX_BOARD][MAX_BOARD];  /* go board */
extern int           ko_i;
extern int           ko_j;
extern int           black_captured;   /* num. of black stones captured */
extern int           white_captured;

/* Reading parameters */
extern int depth;		/* deep reading cutoff */
extern int backfill_depth;	/* deep reading cutoff */
extern int fourlib_depth;	/* deep reading cutoff */
extern int ko_depth;		/* deep ko reading cutoff */

extern int stackp;            /* stack pointer */

/* FIXME: perhaps some of the following should be flag-bits in debug */
extern int count_variations;      /* count (decidestring) */
extern int sgf_dump;              /* writing file (decidestring) */


struct stats_data {
  int  nodes;			  /* Number of visited nodes while reading */
  int  position_entered;	  /* Number of Positions entered. */
  int  position_hits;		  /* Number of hits of Positions. */
  int  read_result_entered;	  /* Number of Read_results entered. */
  int  read_result_hits;	  /* Number of hits of Read_results. */
  int  hash_collisions;           /* Number of hash collisions. */
};

extern struct stats_data stats;


struct half_eye_data {
  int type;         /* HALF_EYE or FALSE_EYE; */
  int num_attacks;  /* number of attacking points */
  int num_defends;  /* number of defending points */
  int ai[4];        /* (ai, aj) is the move to attack a topological halfeye */
  int aj[4];
  int di[4];        /* (di, dj) is the move to defend a topological halfeye */
  int dj[4];
};

extern struct half_eye_data half_eye[MAX_BOARD][MAX_BOARD];      /* array of half-eye data */

/*
 * data concerning a worm. A copy is kept at each vertex of the worm.
 */

struct worm_data {
  int color;         /* its color */
  int size;          /* its cardinality */
  float effective_size; /* stones and surrounding spaces */
  int origini;       /* (origini,originj) is the origin of the string. Two */
  int originj;       /* vertices are in the same worm iff they have same origin. */
  int liberties;     /* number of liberties */
  int liberties2;    /* number of second order liberties */
  int liberties3;    /* third order liberties (empty vertices at distance 3) */
  int liberties4;    /* fourth order liberties */
  int attacki;       /* (attacki, attackj), if attacki != -1 points to move which captures */
  int attackj;       /* the string */
  int attack_code;   /* 1=unconditional win, 2 or 3=win with ko */
  int lunchi;        /* if lunchi != -1 then (lunchi,lunchj) points to a boundary worm */
  int lunchj;        /* which can be captured easily. */
  int defendi;       /* (defendi, defendj), if defendi != -1 points to a defensive move */
  int defendj;       /* protecting the string */
  int defend_code;   /* 1=unconditional win, 2 or 3=win with ko */
  int cutstone;      /* 1=potential cutting stone; 2=cutting stone */
  int cutstone2;     /* Number of potential cuts involving the worm. */
  int genus;         /* number of connected components of the complement, less one */
  int ko;            /* 1=ko at this location */
  int inessential;   /* 1=inessential worm */
};

extern struct worm_data worm[MAX_BOARD][MAX_BOARD];


/*
 * data concerning a dragon. A copy is kept at each stone of the string.
 */

struct dragon_data {
  int color;    /* its color                                                 */
  int origini;  /* (origini,originj) is the origin of the string. Two        */
  int originj;  /* vertices are in the same dragon iff they have same origin.*/
  int borderi;  /* if color=BLACK_BORDER or WHITE_BORDER the worm is a cavity*/
  int borderj;  /* surrounded by the BLACK or WHITE dragon with origin       */
                /* (borderi,borderj)                                         */
  int size;     /* size of the dragon                                        */
  float effective_size; /* stones and surrounding spaces                     */
  int heyes;    /* the number of half eyes                                   */
  int heyei;    /* coordinates of a half eye                                 */
  int heyej;
  int genus;    /* the number of eyes (approximately)                        */
  int escape_route; /* the maximum of worm[*][*].liberties4 over worms in    */
                /* the dragon.  FIXME: Is this still true?                   */
  int escape2;  /* A second measure of running potential.                    */
  int lunchi;   /* if lunchi != -1 then (lunchi,lunchj) points to a boundary */
  int lunchj;   /* worm which can be captured easily.                        */
  int status;   /* (ALIVE, DEAD, UNKNOWN, CRITICAL)                          */
  int mouse_status;   /* (ALIVE, DEAD, UNKNOWN, CRITICAL, UNCHECKED)         */
  int mouse_attacki;  /* vital point for attack                              */
  int mouse_attackj;
  int mouse_defendi;  /* vital point for defense                             */
  int mouse_defendj;
  int safety;   /* A modified status taking into account weakness.           */
  int vitality; /* another measure of life potential (for semeai)            */
  int semeai;   /* true if a string is part of a semeai                      */
};

extern struct dragon_data dragon[MAX_BOARD][MAX_BOARD];


struct eye_data {
  int color;/* BLACK, WHITE, BLACK_BORDERED, WHITE_BORDERED or GRAY_BORDERED */
  int esize;      /* size of the eyespace                                 */
  int msize;      /* number of marginal vertices                          */
  int origini;    /* The origin                                           */
  int originj;    /*                                                      */
  int maxeye;     /* number of eyes if defender plays first               */
  int mineye;     /* number of eyes if attacker plays first               */
  int attacki;    /* vital point for attack                               */
  int attackj;    /*                                                      */
  int defendi;    /* vital point for defense                              */
  int defendj;    /*                                                      */
  int dragoni;    /* origin of the surrounding dragon                     */
  int dragonj;
  /* The above fields are constant on the whole eyespace. */
  /* ---------------------------------------------------------------- */
  /* The below fields are not. */
  int marginal;             /* This vertex is marginal                    */
  int false_margin;         /* This vertex is false margin                */
  int type;                 /* Various characteristics of the eyespace    */
  int neighbors;            /* number of neighbors in eyespace            */
  int marginal_neighbors;   /* number of marginal neighbors               */
  int cut;                  /* Opponent can cut at vertex.                */
};

extern struct eye_data white_eye[MAX_BOARD][MAX_BOARD];
extern struct eye_data black_eye[MAX_BOARD][MAX_BOARD];
extern struct eye_data mouse_black_eye[MAX_BOARD][MAX_BOARD];
extern struct eye_data mouse_white_eye[MAX_BOARD][MAX_BOARD];


/* The following declarations have to be postponed until after the
 * definition of struct eye_data.
 */

void compute_eyes(int i, int  j, int *max, int *min,
		  int *attacki, int *attackj, int *defendi, int *defendj,
		  struct eye_data eye[MAX_BOARD][MAX_BOARD],
		  int add_moves, int color);
void recognize_eye2(int m, int n, int *attacki, int *attackj,
		    int *defendi, int *defendj, int *max, int *min,
		    struct eye_data eye[MAX_BOARD][MAX_BOARD],
		    int add_moves, int color);
void propagate_eye (int, int, struct eye_data eye[MAX_BOARD][MAX_BOARD]);
void originate_eye(int i, int j, int m, int n,
			  int *esize, int *msize,
			  struct eye_data eye[MAX_BOARD][MAX_BOARD]);
int topological_eye(int m, int n, int color, int *ai, int *aj,
		    int *di, int *dj, 
		    struct eye_data b_eye[MAX_BOARD][MAX_BOARD],
		    struct eye_data w_eye[MAX_BOARD][MAX_BOARD]);

void add_half_eye(int m, int n, struct eye_data eye[MAX_BOARD][MAX_BOARD], 
		  struct half_eye_data hey[MAX_BOARD][MAX_BOARD]);
void make_domains(struct eye_data b_eye[MAX_BOARD][MAX_BOARD],
		  struct eye_data w_eye[MAX_BOARD][MAX_BOARD]);

extern int distance_to_black[MAX_BOARD][MAX_BOARD];
extern int distance_to_white[MAX_BOARD][MAX_BOARD];
extern int strategic_distance_to_black[MAX_BOARD][MAX_BOARD];
extern int strategic_distance_to_white[MAX_BOARD][MAX_BOARD];
extern int black_domain[MAX_BOARD][MAX_BOARD];
extern int white_domain[MAX_BOARD][MAX_BOARD];


/* our own abort() which prints board state on the way out.
 * i,j is a "relevant" board position for info */
void abortgo(const char *file, int line, const char *msg, int i, int j);

#ifndef NDEBUG
#define ASSERT(x,i,j)  if (x) ; else abortgo(__FILE__, __LINE__, #x,i,j)   /* avoid dangling else */
#else
#define ASSERT(x,i,j)
#endif


#endif  /* _LIBERTY_H_ */


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
