/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/*
 * Location_set: A set of locations on the go board.
 */

typedef struct {
  long  bitmap[MAX_BOARD];
} Location_set;

typedef struct {
  Location_set  * ls;
  int             current_bitmap;
  int             current_bit;
} Location_set_iterator;

void location_set_clear(Location_set *ls);
void location_set_add(Location_set *ls, int m, int n);
void location_set_del(Location_set *ls, int m, int n);
void location_set_print(FILE *outfile, Location_set *ls);
void location_set_dump(FILE *outfile, Location_set *ls);

void location_set_copy(Location_set *to, Location_set *from);
void location_set_expand(Location_set *ls, int diagonally);
void location_set_union(Location_set *ls, Location_set *other);
void location_set_intersection(Location_set *ls, Location_set *other);
void location_set_subtract(Location_set *ls, Location_set *other);

int location_set_isempty(Location_set *ls);


/* 
 * Usage:
 *
 * Location_set            set;
 * Location_set_iterator   iter;
 * int                     x, y;
 *
 * for (ls_iter_init(&set, &iter);
 *      ls_iter_next(&iter, &x, &y);
 *      ls_iter_update(&iter))
 *   use_it(x, y);
 */

void ls_iter_init(Location_set *ls, Location_set_iterator *lsi);
int  ls_iter_next(Location_set_iterator *lsi, int *m, int *n);
void ls_iter_update(Location_set_iterator *lsi);


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
