/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */





/*-------------------------------------------------------------------------
 * interface.h
 * 	This file contains all headers for interfaces
 *-------------------------------------------------------------------------*/

#ifndef _PLAY_INTERFACE_H
#define _PLAY_INTERFACE_H

#include "gnugo.h"
#include "ttsgf.h"

enum testmode {UNKNOWN_TESTMODE=0, MOVE_ONLY, ANNOTATION_ONLY, BOTH, GAME};

void play_ascii(Gameinfo *gameinfo, char *filename, char *until);
void play_ascii_emacs(Gameinfo *gameinfo, char *filename, char *until);
void play_gtp(Gameinfo *gameinfo);
void play_gmp(int boardsize);
void play_solo(int boardsize, int benchmark);
void play_test(SGFNode *sgf_head, enum testmode mode);

int load_sgf_file(SGFNode *head, Gameinfo *gameinfo, const char *untilstr);
void load_sgf_header(SGFNode *head, Gameinfo *gameinfo);
void load_and_analyze_sgf_file(SGFNode *head, Gameinfo *gameinfo,
			       const char *untilstr, int benchmark);
void load_and_score_sgf_file(SGFNode *head, Gameinfo *gameinfo,
			     const char *untilstr);


/* ---------------------------------------------------------------
 * These functions deal with move manipulation
 * --------------------------------------------------------------- */
int get_move(Gameinfo *gameinfo, int *x, int *y, int move);
int get_test_move(Gameinfo *gameinfo, int *x, int *y, int move);
int put_move(Position *pos, int x, int y, int move);
int put_stone(Position *pos, int x, int y, int color);


/* ---------------------------------------------------------------
 * These functions deal with the game option structure
 * --------------------------------------------------------------- */
void init_gopt(void);
int get_opt_quiet(void);
void set_opt_quiet(int num);
int get_opt_display_board(void);
void set_opt_display_board(int num);



#endif


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
